/* 
   Arnie for PalmOS
   (C)2000 Adam Sampson, azz@gnu.org
   Released under the terms of the GNU General Public License,
   version 2 or above.

   Entry 0 is the start of a sentence; entry 1 is the end of a sentence.
   record format: "string\0", <pad> UInt count, <UInt next0 .. UInt nextN>
   <pad> is a single byte to pad the UInt to an even boundary
   
   Note that the 50-char limit in AppStdIO:PrvProcessCommand is probably
   too small for Arnie.
*/

#include <Pilot.h>
#include "arnie.h"
#include <stdarg.h>
#include "AppStdIO.h"

#define E(X) ErrFatalDisplayIf(X, "error");
#define D(X) StdClear();printf(X"\n");

UInt cardnum = 0;

FormPtr conform;
DmOpenRef dbref;

UInt get_db_size() {
  LocalID dbid;
  UInt card;
  ULong dblen;
  
  // Get the database ID and card.
  E(!dbref);
  E(DmOpenDatabaseInfo(dbref, (LocalID *)&dbid, NULL, NULL, &card, NULL));

  // Get the number of items in the database.
  E(DmDatabaseSize(card, dbid, &dblen, NULL, NULL));

  return dblen;
}

UInt fetch_word(CharPtr s) {
  VoidHand rec;
  VoidPtr data;
  UInt pos, i, len, uis;
  ULong dblen;

  dblen = get_db_size();

  // Search through the items for the specified word.
  for (i=0; i<dblen; i++) {
    Int cmp;

    // Compare the record's data with the key.
    rec = DmQueryRecord(dbref, i);
    data = MemHandleLock(rec);
    cmp = StrCompare(data, s);
    MemHandleUnlock(rec);

    // If it was found, return it.
    if (!cmp) return i;
  }

  // The item wasn't found; we need to add it.
  pos = dmMaxRecordIndex;
  uis = sizeof(UInt);
  len = StrLen(s) + 1;
  if (len%2) len++;
  rec = DmNewRecord(dbref, &pos, len + uis);
  data = MemHandleLock(rec);
  DmStrCopy(data, 0, s);
  i = 0;
  E(DmWrite(data, len, &i, uis));
  MemHandleUnlock(rec);
  E(DmReleaseRecord(dbref, pos, true));
  return pos;
}

void add_link(UInt from, UInt to) {
  VoidHand rec;
  VoidPtr data;
  UInt size, newto, uis, count, countpos;

  uis = sizeof(UInt);
  newto = to;

  rec = DmQueryRecord(dbref, from); E(!rec);
  data = MemHandleLock(rec); E(!data);
  size = StrLen(data) + 1;
  if (size%2) size++;
  data += size;
  count = *(UInt *)data;
  countpos = size;
  size += uis * (++count);
  E(MemHandleUnlock(rec));

  size += uis;
  rec = DmResizeRecord(dbref, from, size);
  E(DmWrite(data = MemHandleLock(rec), size - uis, &newto, uis));
  E(DmWrite(data, countpos, &count, uis));
  E(MemHandleUnlock(rec));
  E(DmReleaseRecord(dbref, from, true));
}

// Say something.
void speak() {
  UInt w, len;

  printf("Arnie> ");

  w = 0;
  do {
    VoidHand rec;
    VoidPtr data;

    rec = DmQueryRecord(dbref, w);
    data = MemHandleLock(rec);

    // Print the string and advance the pointer to point
    // to the count of "next" values.
    if (w) printf("%s ", data);
    len = StrLen(data) + 1;
    data += len;
    if (len%2) data++;

    // Pick a random next word to go to.
    if (*(UIntPtr)data) {
      w = ((UIntPtr)data)[(SysRandom(0)%(*(UIntPtr)data)) + 1];
    } else {
      // No next data? Hmmm.
      printf("no next data at word %d\n", w);
      w = 1;
    }
    MemHandleUnlock(rec);
  } while (w != 1);
  printf("\n");
}

// This trashes the string sent in, which is OK
// because AppStdIO copies it.
void ingest(CharPtr sent) {
  CharPtr spc;
  Boolean onlast = false;
  UInt w, pw;

  pw = 0;
  do {
    // Find the next space or null.
    spc = sent;
    while (*spc && *spc != ' ') spc++;
    if (!*spc) onlast = true;
    *spc = '\0';

    w = fetch_word(sent);
    add_link(pw, w);

    sent = spc + 1;
    pw = w;
  } while (!onlast);
  add_link(w, 1);
}

// Process input from the console.
void HandleInput(CharPtr command) {
  ingest(command);
  speak();
}

void open_database() {
  // Attempt to get the database open; create it if it doesn't exist.
  dbref = DmOpenDatabaseByTypeCreator('Data', APPID, dmModeReadWrite);
  if (!dbref) {
    Char s[80];
    
    E(DmCreateDatabase(cardnum, "ArnieBrain", APPID, 'Data', false));
    dbref = DmOpenDatabaseByTypeCreator('Data', APPID, dmModeReadWrite);
    fetch_word("<start>");
    fetch_word("<end>");
    StrCopy(s, "hello, world");
    ingest(s);
  }
  E(!dbref);

  printf("Welcome to Arnie. The dictionary contains %d words.\n", 
	 get_db_size());
}

static Boolean ConsoleHandleEvent(EventPtr e) {
  Word field;

  // Get the console to handle its events first.
  if (StdHandleEvent(e)) return true;

  switch (e->eType) {
  case frmOpenEvent:

    conform = FrmGetActiveForm();
    FrmEraseForm(conform);
    FrmUpdateForm(ID_conform, frmRedrawUpdateCode);

    return true;
    break;

  case ctlSelectEvent:
    switch(e->data.ctlSelect.controlID) {
    case ID_conclear:
      StdClear();
      return true;

    case ID_conspeak:
      speak();
      return true;
      
    }
    break;
    
  case menuEvent:
    MenuEraseStatus(NULL);
    field = FrmGetFocus(conform);
    switch (e->data.menu.itemID) {
    case ID_confilewipe:
      {
	LocalID dbid;

	dbid = DmFindDatabase(cardnum, "ArnieBrain");
	if (dbid) E(DmDeleteDatabase(cardnum, dbid));

	open_database();
      }
      break;

    case ID_coneditcopy:
      FldCopy(FrmGetObjectPtr(conform, field));
      break;
      
    case ID_coneditpaste:
      FldPaste(FrmGetObjectPtr(conform, field));
      break;

    case ID_conhelpabout:
      FrmAlert(ID_about);
      return true;
      
    }

  }

  return false;
}

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
  EventType e;
  Word error;

  if (cmd == sysAppLaunchCmdNormalLaunch) {
    
    // Initialise the console.
    StdInit(ID_conform, ID_confield, ID_conscroll, HandleInput);

    // Open the database.
    open_database();

    // Switch to the main form.
    FrmGotoForm(ID_conform);

    /* Event loop */
    do {
      EvtGetEvent(&e, evtWaitForever);
      if (SysHandleEvent(&e)) continue;
      if (MenuHandleEvent(NULL, &e, &error)) continue;

      if (e.eType == frmLoadEvent) {
	FormPtr frm = FrmInitForm(e.data.frmLoad.formID);
	FrmSetActiveForm(frm);
	switch (e.data.frmLoad.formID) {
	case ID_conform:
	  FrmSetEventHandler(frm, ConsoleHandleEvent);
	  break;
	}
	continue;
      }

      FrmDispatchEvent(&e);

    } while (e.eType != appStopEvent);

    // Close the database.
    DmCloseDatabase(dbref);

    // Free the console
    StdFree();

  } else {
    return sysErrParamErr;
  }
  return 0;
}




